/**
 * Easy access to RS-232 port.
 * This library uses only POSIX calls and no direct hardware access.
 *
 * Lars Stollenwerk, 2010
 *
 * use with:
 * close(2), read(2), write(2)
 */
#ifndef LIB232_H
#define LIB232_H

#include <stdio.h>
#include <termios.h>
#include <unistd.h>

/**
 * Open a serial port.
 * The outgoing chars are always sent raw. All echoing is switched off.
 *
 * tty    device of serial port, e.g. "/dev/ttyS0"
 * baud   baud rate
 * frame  string comprising 
 *        byte lenth (5,6,7,8), parity (n,o,e), stop bits (1,2); e.g.
 *        "8n1"
 * hs     handshake: NONE, XONXOFF, RTSCTS
 * mode   CANON (read linewise, line ends on NL or eof, CR is
 *        translated to NL);
 *        RAW  (read raw bytes)
 *
 * return: file descriptor or -1 on error (see errno)
 */
int open232( const char * tty, int baud, const char * frame, int hs, int mode );

/* constants for open232 */
#define NONE    0
#define XONXOFF 1
#define RTSCTS  2
#define RAW     0
#define CANON   1

/**
 * Print status of serial handshake lines to stderr:
 * DTR, DSR, RTS, CTS, CD, RI
 *
 * fd      file descriptor of serial port
 * return: 0 on success; -1 on error, see errno
 */
int print_rs232_state( int fd );

/**
 * Send break.
 * This is useful to reset the line when the communication crashed.
 *
 * fd      file descriptor of serial port
 * return: 0 on success; -1 on error, see errno
 */
#define break232(x) tcsendbreak((x),0)

#endif
