/**
 * osc_api -  library to communicate with HAMEG digital storage oscilloscopes
 *            via serial port.
 *         
 * Copyright (C) 2007, 2015 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * formica@opppf.de
 * http://opppf.de
 */
#include "osc_api.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>


/* debugging */
#define debugc(x) fprintf( stderr, "%s=%c %02x\n", #x, (x), (unsigned int)(x) )
#define debugi(x) fprintf( stderr, "%s=%d %04x\n", #x, (x), (unsigned int)(x) )
#define debugd(x) fprintf( stderr, "%s=%f\n", #x, (x) )

/* ----------------------------------------------------------------------
			      internal functions
 ---------------------------------------------------------------------- */

/**
 * Rerset error indicator. 
 * Must be called by every function that sets error[].
 */
#define clear_error() o->error[0]=0


/** 
 * Return on error
 * Call to give error to upper instance.
 */
#define on_error_return(x) if(o->error[0])return(x)


/** 
 * timeout for osc_read_bytes() in 100 ms.
 */
#define TIMEOUT 80


/**
 * Read n bytes from scope.
 *
 * o   OSC struct
 * b   buffer to write to
 * m   number characters to read
 * return: 0 = success, -1 = error
 */
int osc_read_bytes( OSC * o, char * b, int m )
{
  fd_set readfds;
  struct timeval timeout;
  int n, i, selected;

  clear_error();
  n = 0; /* bytes read */
  for ( i = 0; i < TIMEOUT; i++ )
  {
    FD_ZERO( &readfds );
    FD_SET( o->fd, &readfds ); 
    timeout.tv_sec = 0;
    timeout.tv_usec = 100000; /* 100 ms */
    selected = select( o->fd + 1, &readfds, NULL, NULL, &timeout );
    switch ( selected )
    {
      case -1:
	strcpy( o->error, "Something wrong with select()." );
	OSC_SHOUT;
	return -1;
      case 0:
        /* just wait */
        break;
      case 1:
	n += read( o->fd, &b[n], m-n ); 
	if ( n == m )
	  return 0;
	i = 0; /* reset timer for timeout */
	break;
      default:
	sprintf( o->error, "select() returned %d.", selected );
	OSC_SHOUT;
	return -1;
    }
  }
  strcpy( o->error, "osc_read_bytes() reached timeout." );
  OSC_SHOUT;
  return -1;
}


/**
 * Calculate high and low byte from short integer.
 * works for signed and unsigned type.
 *
 * i	integer
 * l	low byte
 * h	high byte
 */
void osc_lo_hi_byte( const void * i, unsigned char * l, unsigned char * h )
{
  const unsigned short int * I = (unsigned short int*)i;

  *h = *I / 256;
  *l = *I - *h * 256;
}


/* ----------------------------------------------------------------------
			     low level functions
 Basic communication. 
 Returncode must be checked.
 ---------------------------------------------------------------------- */


/**
 * Open connection to scope.
 *
 * c	   filename of serial port
 * return: OSC descriptor
 *         possible error in field error.
 */
OSC * osc_open( const char * c )
{
  OSC * o;
  int e;
  struct termios options;
  
  /* init */
  o = malloc( sizeof(OSC) );
  o->tty = malloc( (1 + strlen(c)) * sizeof(char) ); 
  strcpy( o->tty, c );
  clear_error();
  /* open port */
  o->fd = open( c, O_RDWR | O_NOCTTY );
  if ( o->fd == -1 ) 
  { 
    strcpy( o->error, "Can't open serial port." ); 
    OSC_SHOUT;
    return o; 
  }
  /* get port options */
  if ( tcgetattr( o->fd, &options ) == -1 )
  { 
    strcpy( o->error, "Can't access serial port." ); 
    OSC_SHOUT;
    close( o->fd );
    return o; 
  }
  /* baud rate */
  e = 0;
  e += cfsetispeed( &options, B19200 );
  e += cfsetospeed( &options, B19200 );
  /* select 8N2 */
  options.c_cflag &= ~CSIZE; 
  options.c_cflag |= CS8;   
  options.c_cflag |= CSTOPB;   
  options.c_cflag &= ~PARENB;
  /* RTS/CTS handshake */
  options.c_cflag |= CRTSCTS;   
  /* raw in- and output */
  options.c_iflag |= IGNBRK | IGNPAR;
  options.c_iflag &= ~(BRKINT|PARMRK|INPCK|ISTRIP|INLCR|IGNCR|ICRNL|IXON|IXOFF);
  options.c_oflag &= ~( OPOST | ONLCR | OCRNL | ONOCR | ONLRET | OFILL );
  options.c_cflag |= ( CLOCAL | CREAD | HUPCL );
  options.c_lflag &= ~( ICANON | ECHO | ECHOE | ISIG );
  /* set port options */
  e += tcsetattr( o->fd, TCSANOW, &options );
  /* init scope: send SPACE CR, get returncode */
  if ( write( o->fd, " \r", 2 ) != 2 ) { e += -1; }
  if ( e != 0 ) 
  { 
    strcpy( o->error, "Can't set up serial port." ); 
    OSC_SHOUT;
    close( o->fd );
    return o; 
  }
  osc_check_returncode( o, "osc_open" ); on_error_return( o );
  /* allow user interaction with autoset key */
  osc_write_str( o, "LK=1" ); on_error_return( o );
  osc_check_returncode( o, "osc_open" ); on_error_return( o );
  /* determine hardware */
  osc_write_str( o, "ID?" ); on_error_return( o );
  osc_read_str( o, "ID:", o->id, 27 ); on_error_return( o );
  o->id[25] = 0;
  o->model = atoi( &o->id[2] );
  /* determine firmware */
  osc_write_str( o, "VERS?" ); on_error_return( o );
  osc_read_str( o, "VERS:", o->version, 15 ); on_error_return( o );
  o->version[13] = 0;
  o->fc = atof( &o->version[2] );
  /* read scope state */
  osc_read_state( o ); on_error_return( o );
  return o;
}


/**
 * Close Connection to scope.
 *
 * o	OSC struct
 */
void osc_close( OSC * o )
{
  char e[OSC_ERROR_STR_LEN];

  osc_write_str( o, "RM0" );
  strcpy( e, o->error );
  osc_check_returncode( o, "osc_close" );
  strcat( e, o->error );
  strcpy( o->error, e );
  close( o->fd );
  free( o->tty );
  free( o );
}


/**
 * Write 0 terminated string to scope.
 *
 * o	OSC struct
 * c	string to write
 * return: 0 = success; -1 = error
 */
int osc_write_str( OSC * o, const char * c )
{
  int e1, e2;

#ifdef OSC_DEBUG
  fprintf( stderr, "osc_write_str : '%s'\n", c );
#endif
  clear_error();
  e1 = write( o->fd, c, strlen( c ) );
  e2 = write( o->fd, "\r\n", 2 );
  if ( e1 == -1 || e2 == -1 )
  {
    sprintf( o->error, "osc_write_str(): error writing %s.", c ); 
    OSC_SHOUT;
    return -1;
  }
  return 0;
}


/**
 * Write string with n bytes to scope.
 *
 * o	OSC struct
 * c	string to write
 * n	length of string
 * return: 0 = success; -1 = error
 */
int osc_write_strn( OSC * o, const char * c, int n )
{
  int e1, e2;

  clear_error();
#ifdef OSC_DEBUG
  fputs( "osc_write_strn: '", stderr );
  fwrite( c, sizeof(char), n, stderr );
  fputs( "'\n", stderr );
#endif
  e1 = write( o->fd, c, n );
  e2 = write( o->fd, "\r\n", 2 );
  if ( e1 == -1 || e2 == -1 )
  {
    strcpy( o->error, "Error in osc_write_strn()." ); 
    OSC_SHOUT;
    return -1;
  }
  return 0;
}


/**
 * Read answer from scope.
 * Answer comprises string c and m further bytes, written to b.
 *
 * o	OSC struct
 * c	string to receive
 * b	destination for answer
 * m	bytes of answer
 * return: 0 = success; -1 = error
 */
int osc_read_str( OSC * o, const char * c, void * b, int m )
{
  int n;
  char cb[256];

  clear_error();
  /* read string part of answer */
  n = strlen( c );
  osc_read_bytes( o, cb, n ); on_error_return( -1 );
  cb[n] = 0;
#ifdef OSC_DEBUG
  fprintf( stderr, "osc_read_str  : '%s'\n", cb );
#endif
  if ( strcmp( cb, c ) != 0 )
  {
    sprintf( o->error, "osc_read_str(): error reading %s.", c ); 
    OSC_SHOUT;
    return -1;
  }
  /* read data part of answer */
  osc_read_bytes( o, b, m ); on_error_return( -1 );
  return 0;
}


/**
 * Read answer from scope.
 * Answer comprises string c, length n, and m further bytes, written to b.
 *
 * o	OSC struct
 * c	string to receive
 * b	destination for answer
 * m	bytes of answer
 * return: 0 = success; -1 = error
 */
int osc_read_strn( OSC * o, const char * c, int n, void * b, int m )
{
  char cb[256];

  /* clear errror message */
  clear_error();
  /* read string part of answer */
  osc_read_bytes( o, cb, n ); on_error_return( -1 );
#ifdef OSC_DEBUG
  fputs( "osc_read_strn : '", stderr );
  fwrite( cb, sizeof(char), n, stderr );
  fputs( "'\n", stderr );
#endif
  if ( memcmp( cb, c, n ) != 0 )
  {
    strcpy( o->error, "osc_read_strn" ); 
    OSC_SHOUT;
    return -1;
  }
  /* read data part of answer */
  osc_read_bytes( o, b, m ); on_error_return( -1 );
  return 0;
}


/**
 * Read returncode R CR NL
 *
 * o	OSC struct
 * return: The R, -1 on error
 */
int osc_get_returncode( OSC * o )
{
  char b[3];

  clear_error();
  osc_read_bytes( o, b, 3 ); on_error_return( -1 );
#ifdef OSC_DEBUG
  fprintf( stderr, "got returncode: '%c' (%#02x %#02x %#02x)\n", 
           b[0], b[0], b[1], b[2] );
#endif
  /* Caution! No processing of CR and NL by serial kernal driver allowed. */
  if ( b[1] == '\r' && b[2] == '\n' ) 
    return b[0];
  strcpy( o->error, "Can not read return code."  );
  OSC_SHOUT;
  return -1;
}



/**
 * Check whether returncode is '0' = ok.
 * A correct communication with returncode != '0' is considered as error.
 * 
 * s	additional text, e. g. function name.
 * o	OSC struct
 * return: 0  = success, -1 = error;
 */
int osc_check_returncode( OSC * o, const char * s )
{
  char b[3];

  clear_error();
  osc_read_bytes( o, b, 3 ); on_error_return( -1 );
#ifdef OSC_DEBUG
  fprintf( stderr, "got returncode: '%c' (%#02x %#02x %#02x)\n", 
           b[0], b[0], b[1], b[2] );
#endif
  /* Caution! No processing of CR and NL by serial kernal driver allowed. */
  if ( b[1] != '\r' || b[2] != '\n' ) 
  {
    strcpy( o->error, "Can not read return code."  );
    OSC_SHOUT;
    return -1;
  }
  switch ( b[0] )
  {
    case '0': return 0;
    case '1': sprintf( o->error, "%s: Returncode 1: syntax error.", s );
	      OSC_SHOUT;
              return -1;
    case '2': sprintf( o->error, "%s: Returncode 2: data error.", s );
	      OSC_SHOUT;
              return -1;
    case '3': sprintf( o->error, "%s: Returncode 3: buffer overflow.", s );
	      OSC_SHOUT;
              return -1;
    case '4': sprintf( o->error, "%s: Returncode 4: bad data set.", s );
	      OSC_SHOUT;
              return -1;
    case '5': sprintf( o->error, "%s: Returncode 5: adjustment error.", s );
	      OSC_SHOUT;
              return -1;
    case '6': sprintf( o->error, "%s: Returncode 6: timing error.", s );
	      OSC_SHOUT;
              return -1;
    default: sprintf( o->error, "%s: Returncode %#04x unknown.", 
                     s, (unsigned int)b[0] );
	     OSC_SHOUT;
             return -1;
  }
}


/* ----------------------------------------------------------------------
			     high level functions
 Complete communication for a task.
 No further check of retruncode needed. 
 ---------------------------------------------------------------------- */


/**
 * Read scope state. 
 * Is automaticcaly done on open and set_auto.
 *
 * o	OSC struct
 * return: 0 = success; -1 = error
 */
int osc_read_state( OSC * o )
{
  // todo: check endianess
  osc_write_str( o, "DDF?" ); on_error_return( -1 );
  osc_read_str( o, "DDF:", &o->ddf, 14 ); on_error_return( -1 );
  osc_write_str( o, "DDF1?" ); on_error_return( -1 );
  osc_read_str( o, "DDF1:", &o->ddf1, 16 ); on_error_return( -1 );
  return 0;
}


/**
 * Perform autoset.
 *
 * o	OSC struct
 * return: 0 = success; -1 = error
 */
int osc_set_auto( OSC * o )
{
  osc_write_str( o, "AUTOSET" ); on_error_return( -1 );
  osc_check_returncode( o, "osc_set_auto: AUTOSET" ); on_error_return( -1 );
  osc_read_state( o );
  return 0;
}


/**
 * bell
 *
 * o	OSC struct
 * return: 0 = success; -1 = error
 */
int osc_bell( OSC * o, int b )
{
  char s[6];

  clear_error();
  if ( b < 0 || b > 5 )
  {
    sprintf( o->error, "Unknown bell code %d.", b );
    OSC_SHOUT;
    return -1;
  }
  sprintf( s, "BELL=%c", b );
  osc_write_str( o, s ); on_error_return( -1 );
  osc_check_returncode( o, "osc_bell: BELL" ); on_error_return( -1 );
  return 0;
}


/**
 * Was scope used manually?
 *
 * o    	OSC struct
 * return: 1 = yes, 0 = no, -1 = error
 */
int osc_changed_manually( OSC * o )
{
  char c;

  clear_error();
  if ( o->fc < 1.1 )
  {
    sprintf( o->error, "osc_changed_manually (FCCMD) needs FC 1.1; DC is %f.",
             o->fc );
    OSC_SHOUT;
    return -1;
  }
  osc_write_str( o, "FCCMD?" ); on_error_return( -1 );
  osc_read_str( o, "FCCMD:", &c, 1 ); on_error_return( -1 );
  switch ( c )
  {
    case '0': return 0;
    case '1': return 1;
    default:
      sprintf( o->error, "osc_changed_manually got illegal answer %#04x.", 
               (unsigned int)c );
      OSC_SHOUT;
      return -1;
  }
}

/* --- data acquisition ------------------------------------------------- */

/**
 * Set stormode.
 * mode is one of OSC_STR_MODE_..  SGL, ROL, ENV, AVM
 *
 * o            scope handle
 * mode		desired mode
 * return: 0 = success, -1 = error
 */
int osc_set_stormode( OSC * o, int mode )
{
  char s[10];

  clear_error();
  if ( mode & ~OSC_STR_MODE )
  {
    sprintf( o->error, "Illegal store mode %d.", mode );
    OSC_SHOUT;
    return -1;
  }
  o->ddf.strmode &= ~OSC_STR_MODE;
  o->ddf.strmode |= (ddft)mode;
  sprintf( s, "STRMODE=%c", o->ddf.strmode );
  osc_write_strn( o, s, 9 ); on_error_return( -1 );
  osc_check_returncode( o, "osc_set_stormode: STRMODE" ); on_error_return( -1 );
  return 0;
}


/**
 * Read data from cahannel n.
 * Values in Volt.
 * Read beginns at offset os and is len bytes long.
 * read whole channel: os = 0, len = 2048
 *
 * o		OSC struct
 * ch		channel number 1, 2
 * os		offset
 * len		length
 * b		buffer to write to, length len
 * return: 0 = success; -1 = error; -2 = clip
 */
int osc_read_chan( OSC * o, int ch, int os, int len, double * b )
{
  int ret, i;
  char s[13];
  unsigned char lo, ho, ll, hl;
  unsigned char buf[2048];
  double Vdiv;

  clear_error();
  ret = 0;
  if ( ch != 1 && ch != 2 )
  {
    sprintf( o->error, "Illegal channel %d.", ch );
    OSC_SHOUT;
    return -1;
  }
  if ( os < 0 || len < 0 || os + len > 2048 )
  {
    strcpy( o->error, "Range must be in [0..2048]." );
    OSC_SHOUT;
    return -1;
  }
  /* prepare command */
  osc_lo_hi_byte( &os, &lo, &ho );
  osc_lo_hi_byte( &len, &ll, &hl );
  sprintf( s, "RDWFM%d:%c%c%s%c%c%s", ch, lo, ho,         /* old FC needs ',' */
                                     (o->fc < 1.05)?",":"", 
				     ll, hl,             /* old FC needs ':' */
				     (o->fc < 1.05)?":":"" );
  /* read data *        v___don't_send_':'___v  */
  osc_write_strn( o, s, (o->fc < 1.05)?12:11 ); on_error_return( -1 );
  osc_read_strn( o, s, (o->fc < 1.05)?13:11, buf, len ); on_error_return( -1 );
  /* transfer to real voltage */
  Vdiv = OSC_V_DIV[ (ch==1 ? o->ddf.ch1 : o->ddf.ch2) & OSC_CHx_YS ];
  if ( o->fc >= 2.0 )
    if ( o->ddf.vermode & (ch==1?OSC_VER_TK1:OSC_VER_TK2) )
      Vdiv *= 10; /* probe x10 */
  for ( i = 0; i < len; i++ )
  {
    if ( buf[i] == 0x00 || buf[i] == 0xff )
	ret = -2; /* clip */
    b[i] = ( buf[i] - 127 ) / 25.0 * Vdiv;
  }
  return ret;
}


/**
 * set average number for acquisition
 *
 * o	OSC struct
 * e	exponent: number is 2^e; e in [1..9]
 * return: 0 = success; -1 = error
 */
int osc_set_avrnm( OSC * o, int e )
{
  clear_error();
  if ( e < 0 || e > 9 )
  {
    strcpy( o->error, "Exponent must be in [1..9]." );
    OSC_SHOUT;
    return -1;
  }
  osc_write_strn( o, "AVRNM=%c", e ); on_error_return( -1 );
  osc_check_returncode( o, "osc_set_avrnm: AVRNM" ); on_error_return( -1 );
  return 0;
}


/**
 * get average number for acquisition.
 * number is 2^e; e in [0..9].
 *
 * o	OSC struct
 * return: e, -1 = error
 */
int osc_get_avrnm( OSC * o )
/* This is not part of DDF ??? */
{
  char c;

  osc_write_str( o, "AVRM?" ); on_error_return( -1 );
  osc_read_str( o, "AVRM:", &c, 1 ); on_error_return( -1 );
  return c;
}


/* --- Y settings ------------------------------------------------------- */

/**
 * Set channel settings.
 * state is a sum of OSC_CHx_GND, OSC_CHx_AC, OSC_CHx_INV, OSC_CHx_ON
 *
 * o    	OSC struct
 * ch		channel, 1 or 2
 * state	desired state, -1 = don't change
 * ys		y-scale as index in OSC_V_DIV, -1 = don't change
 * var  	varscale, 0..0xff, 0xff = claibrated, -1 = don't change
 * return: 0 = success; -1 = error
 */
int osc_set_ch( OSC * o, int ch, int state, int ys, int var )
{
  char s[9];
  ddft * ddf_chx;
  ddft * chXvar;

  clear_error();
  /* check range */
  if ( ch != 1 && ch != 2 )
  {
    sprintf( o->error, "Illegal channel %d.", ch );
    OSC_SHOUT;
    return -1;
  }
  if ( state > 0 && ( state & OSC_CHx_YS || state > 0xf0 ) )
  {
    sprintf( o->error, "Illegal state %#04x", state );
    OSC_SHOUT;
    return -1;
  }
  if ( ys >= OSC_V_DIVN )
  {
    sprintf( o->error, "Voltage scale %#04x out of range.", ys );
    OSC_SHOUT;
    return -1;
  }
  if ( var > 0 && var > 0xff )
  {
    sprintf( o->error, "Var voltage scale var=%#04x out of range.", var );
    OSC_SHOUT;
    return -1;
  }
  /* work on desired channel */
  if ( ch == 1 )
  {
    ddf_chx = &o->ddf.ch1;
    chXvar = &o->ddf.ch1var;
  }
  else
  {
    ddf_chx = &o->ddf.ch2;
    chXvar = &o->ddf.ch2var;
  }
  /* set state and y-scale */
  if ( state >= 0 )
    *ddf_chx = ( *ddf_chx & OSC_CHx_YS ) | (ddft)state;
  if ( ys >= 0 )
    *ddf_chx = ( *ddf_chx & ~OSC_CHx_YS ) | (ddft)ys;
  if ( state >= 0 || ys >= 0 )
  {
    sprintf( s, "CH%d=%c", ch, *ddf_chx );
    osc_write_strn( o, s, 5 ); on_error_return( -1 );
    osc_check_returncode( o, "osc_set_ch, CHx=b" ); on_error_return( -1 );
  }
  /* set var */
  if ( var >= 0 )
  {
    *chXvar = var;
    sprintf( s, "CH%dVAR=%c", ch, *chXvar );
    osc_write_strn( o, s, 8 ); on_error_return( -1 );
    osc_check_returncode( o, "osc_set_ch, CHxVAR=b" ); on_error_return( -1 );
  }
  return 0;
}


/** Y amplification in DDF to voltage scale in V/DIV */
const double OSC_V_DIV[OSC_V_DIVN] =
{  1e-3,   2e-3,   5e-3, 
  10e-3,  20e-3,  50e-3, 
 100e-3, 200e-3, 500e-3, 
   1,      2,      5, 
  10,     20 };


/**
 * voltage scale to index in OSC_V_DIV
 *
 * vs	voltage scale in V/DIV
 * return: corresponding index in OSC_V_DIV, -1 on error.
 */
int osc_vs_to_int( double vs )
{
  int i;

  for ( i = 0; i < OSC_V_DIVN; i++ )
    if ( OSC_V_DIV[i] == vs )
      return i;
  return -1;
}


/**
 * Get channel state
 * state is a sum of OSC_CHx_GND, OSC_CHx_AC, OSC_CHx_INV, OSC_CHx_ON
 *
 * o    	OSC struct
 * ch		channel, 1 or 2
 * return: state
 */
int osc_get_ch_state( OSC * o, int ch )
{
  switch ( ch )
  {
    case 1:
      return o->ddf.ch1 & ~OSC_CHx_YS;
    case 2:
      return o->ddf.ch2 & ~OSC_CHx_YS;
    default:
      sprintf( o->error, "Illegal channel %d.", ch );
      OSC_SHOUT;
      return -1;
   }
}


/**
 * Get voltage scale.
 *
 * o    	OSC struct
 * var		if not NULL, var setting is written here
 * ch		channel, 1 or 2
 * return: voltage scale as index in OSC_V_DIV
 */
int osc_get_ch_scale( OSC * o, int ch, int * var )
{
  switch ( ch )
  {
    case 1:
      if ( var )
        *var = o->ddf.ch1var;
      return o->ddf.ch1 & OSC_CHx_YS;
    case 2:
      if ( var )
        *var = o->ddf.ch2var;
      return o->ddf.ch2 & OSC_CHx_YS;
    default:
      sprintf( o->error, "Illegal channel %d.", ch );
      OSC_SHOUT;
      return -1;
   }
}


/**
 * Set y position on screen
 * The value is an approximation. The middle may differ for both channels
 * and may be approx. 1 DIV off the screen center.
 *
 * o            OSC struct
 * ch 		channel, 1 or 2
 * pos		-1 = bottom, 0 = middle, 1 = top
 * return: 0 = success, -1 = error
 */
int osc_set_ypos( OSC * o, int ch, double pos )
{
  char s[8];
  unsigned char lo, hi;
  short int * ypos;

  clear_error();
  switch ( ch )
  {
    case 1:
      ypos = &o->ddf1.y1pos;
      break;
    case 2:
      ypos = &o->ddf1.y2pos;
      break;
    default:
      sprintf( o->error, "Illegal channel %d.", ch );
      OSC_SHOUT;
      return -1;
  }
  if ( fabs( pos ) > 1 )
  {
    sprintf( o->error, "YPos %f not in [-1..1].", pos );
    OSC_SHOUT;
    return -1;
  }
  if ( o->fc < 1.1 )
    *ypos = ( pos + 1 ) / 2 * 0x3ff;
  else
    *ypos = pos * 4 * 1000;
  osc_lo_hi_byte( ypos, &lo, &hi );
  sprintf( s, "Y%dPOS=%c%c", ch, lo, hi );
  osc_write_strn( o, s, 7 ); on_error_return( -1 );
  osc_check_returncode( o, "osc_set_ypos: YxPOS" ); on_error_return( -1 );
  return 0;
}


/**
 * Get y position on screen
 *
 * o            OSC struct
 * ch 		channel, 1 or 2
 * return: -1 = bottom, 0 = middle, 1 = top
 */
double osc_get_ypos( OSC * o, int ch )
{
  short int * ypos;

  clear_error();
  switch ( ch )
  {
    case 1:
      ypos = &o->ddf1.y1pos;
      break;
    case 2:
      ypos = &o->ddf1.y2pos;
      break;
    default:
      sprintf( o->error, "Illegal channel %d.", ch );
      OSC_SHOUT;
      return -1;
  }
  if ( o->fc < 1.1 )
    return 2 * *ypos / (double)0x3ff - 1;
  else
    return *ypos / 4000.0;
}


/* --- horizontal mode -------------------------------------------------- */

/**
 * Set horizontal mode.
 * Mode comprises OSC_HOR_... CT, XY, x10, STORE, PP, TB_a, TB_ab, TB_alt, TB_b
 * Mask comprises OSC_HOR_... CT, XY, x10, STORE, PP, TB
 * Only features named in mask are changed. They get the value given in mode.
 *
 * o    	OSC struct
 * mode		desired mode
 * mask		desired mask
 * return: 0 = success, -1 = error
 */
int osc_set_hormode( OSC * o, int mode, int mask )
{
  char s[10];

  if ( mode < 0 || mode > 0xff )
  {
    sprintf( o->error, "Illegal hormode %#04x.", mode );
    OSC_SHOUT;
    return -1;
  } 
  if ( mask < 0 || mask > 0xff )
  {
    sprintf( o->error, "Illegal hormode mask %#04x.", mask );
    OSC_SHOUT;
    return -1;
  } 
  o->ddf.hormode &= ~(ddft)mask;
  o->ddf.hormode |= (ddft)mask & (ddft)mode; 
  sprintf( s, "HORMODE=%c", o->ddf.hormode );
  osc_write_strn( o, s, 9 ); on_error_return( -1 );
  osc_check_returncode( o, "osc_set_hormode: HORMODE" ); on_error_return( -1 );
  return 0;
}


/* --- timebase settings ------------------------------------------------ */

/**
 * Set timebase.
 * state for TB A is sum of: OSC_TBA_ZIN, OSC_TBA_ANA
 * state for TB B is sum of: OSC_TBB_NFL, OSC_TBB_BTR
 *
 * o		OSC struct
 * tb		timebas, 'A' or 'B'
 * state	desired state, -1 = don't schange
 * ts		timescale as index in OSC_s_DIV, -1 = don't change
 * var		varscale, 0..0x3ff, 0 = claibrated, -1 = dont't change
 * return: 0 = success; -1 = error
 */
int osc_set_tb( OSC * o, int tb, int state, int ts, int var )
{
  char s[10];
  unsigned char lb, hb;
  ddft * ddf_tbx;
  short int * tbXvar;

  clear_error();
  if ( tb != 'A' && tb != 'B' )
  {
    sprintf( o->error, "Timebase must be 'A' or 'B, not '%c'.", tb );
    OSC_SHOUT;
    return -1;
  }
  if ( state > 0 && ( state & OSC_TBx_TS && state > 0xe0 ) )
  {
    sprintf( o->error, "Illegal state %#04x for TB.", state );
    OSC_SHOUT;
    return -1;
  }
  if ( ts >= OSC_s_DIVN )
  {
    sprintf( o->error, "Timescale %#04x out of range.", ts );
    OSC_SHOUT;
    return -1;
  }
  if ( var > 0x3ff )
  {
    sprintf( o->error, "Timescale var=%#04x out of range.", var );
    OSC_SHOUT;
    return -1;
  }
  /* operate on desired timebase */
  if ( tb == 'A' )
  {
    ddf_tbx = &o->ddf.tba;
    tbXvar = &o->ddf1.tb_a_var;
  }
  else
  {
    ddf_tbx = &o->ddf.tbb;
    tbXvar = &o->ddf1.tb_b_var;
  }
  /* set timebase */
  if ( state >= 0 )
    *ddf_tbx = ( *ddf_tbx & OSC_TBx_TS ) | (ddft)state;
  if ( ts >= 0 )
    *ddf_tbx = ( *ddf_tbx & ~OSC_TBx_TS ) | (ddft)ts;
  if ( state >= 0 || tb >= 0 )
  {
    sprintf( s, "TB%c=%c", tb, *ddf_tbx );
    osc_write_strn( o, s, 5 ); on_error_return( -1 );
    osc_check_returncode( o, "osc_set_tb: TBx=b" ); on_error_return( -1 );
  }
  /* set var */
  if ( var > 0 )
  {
    *tbXvar = var;
    hb = var / 256;
    lb = var - 256 * hb;
    sprintf( s, "TB%cVAR=%c%c", tb, lb, hb );
    osc_write_strn( o, s, 9 ); on_error_return( -1 );
    osc_check_returncode( o, "osc_set_tb: TBxVAR=b" ); on_error_return( -1 );
  }
  return 0;
}



/** Timebase in DDF to timescale in s/DIV  */
const double OSC_s_DIV[OSC_s_DIVN] =
{                  50e-9, 
  100e-9, 200e-9, 500e-9,
    1e-6,   2e-6,   5e-6,
   10e-6,  20e-6,  50e-6,
  100e-6, 200e-6, 500e-6,
    1e-3,   2e-3,   5e-3,
   10e-3,  20e-3,  50e-3,
  100e-3, 200e-3, 500e-3,
    1   ,   2   ,   5   ,
   10   ,  20   ,  50   ,
  100 };


/**
 * timescale to index in OSC_s_DIV
 *
 * ts	timescale in s/DIV
 * return: corresponding index in OSC_s_DIV, -1 on error.
 */
int osc_ts_to_int( double ts )
{
  int i;

  for ( i = 0; i < OSC_s_DIVN; i++ )
    if ( OSC_s_DIV[i] == ts )
      return i;
  return -1;
}


/**
 * Get timebasestate
 * state is sum of:
 * TB A: OSC_TBA_ZIN, OSC_TBA_ANA
 * TB B: OSC_TBB_NFL, OSC_TBB_BTR
 *
 * o            OSC struct
 * tb		timebase 'A' or 'B'
 * return: state
 */
int osc_get_tb_state( OSC * o, int tb )
{
  clear_error();
  switch ( tb )
  {
    case 'A':
      return o->ddf.tba & ~OSC_TBx_TS;
    case 'B':
      return o->ddf.tbb & ~OSC_TBx_TS;
    default:
      sprintf( o->error, "Timebase must be 'A' or 'B, not '%c'.", tb );
      OSC_SHOUT;
      return -1;
  }
}


/**
 * Get time scale
 *
 * o            OSC struct
 * var          if not NULL, var setting is written here
 * b            timebase 'A' or 'B'
 * return: time scale as index in OSC_s_DIV
 */
int osc_get_tb_scale( OSC * o, int tb, int * var )
{
  clear_error();
  switch ( tb )
  {
    case 'A':
      if ( var )
        *var = o->ddf1.tb_a_var;
      return o->ddf.tba & OSC_TBx_TS;
    case 'B':
      if ( var )
        *var = o->ddf1.tb_b_var;
      return o->ddf.tbb & OSC_TBx_TS;
    default:
      sprintf( o->error, "Timebase must be 'A' or 'B, not '%c'.", tb );
      OSC_SHOUT;
      return -1;
  }
}


/**
 * Set x position on screen
 *
 * pos		-1 = very left, 0 = middle, 1 = very right
 * o            OSC struct
 * return: 0 = success, -1 = error
 */
int osc_set_xpos( OSC * o, double pos )
{
  char s[8];
  unsigned char lo, hi;

  clear_error();
  if ( fabs( pos ) > 1 )
  {
    sprintf( o->error, "XPos %f not in [-1..1].", pos );
    OSC_SHOUT;
    return -1;
  }
  if ( o->fc < 1.1 )
    o->ddf1.xpos = ( pos + 1 ) / 2 * 0x3ff;
  else
    o->ddf1.xpos = pos * 5 * 1000;
  osc_lo_hi_byte( &o->ddf1.xpos, &lo, &hi );
  sprintf( s, "XPOS=%c%c", lo, hi );
  osc_write_strn( o, s, 7 ); on_error_return( -1 );
  osc_check_returncode( o, "osc_set_xpos: XPOS" ); on_error_return( -1 );
  return 0;
}


/**
 * Get x position on screen
 *
 * o            OSC struct
 * return: -1 = very left, 0 = middle, 1 = very right
 */
double osc_get_xpos( OSC * o )
{
  if ( o->fc < 1.1 )
    return 2 * o->ddf1.xpos / (double)0x3ff - 1;
  else
    return o->ddf1.xpos / 5000.0;
}


/* --- vertical mode ---------------------------------------------------- */

/**
 * Set vertical mode.
 * Mode comprises OSC_VER_..  ALTTRG, TK1, BWL, CHOP, ADD, TK2, TRS_CH1,
 *                                                              TRS_CH2, TRS_EXT
 * Mask comprises OSC_VER_..  ALTTRG, TK1, BWL, CHOP, ADD, TK2, TRS
 *
 * o            OSC struct
 * mode         desired mode
 * mask         desired mask
 * return: 0 = success, -1 = error
 */
int osc_set_vermode( OSC * o, int mode, int mask )
{
  char s[10];

  clear_error();
  if ( mode < 0 || mode > 0xff )
  {
    sprintf( o->error, "Illegal vermode %#04x.", mode );
    OSC_SHOUT;
    return -1;
  } 
  if ( mask < 0 || mask > 0xff )
  {
    sprintf( o->error, "Illegal vedmode mask %#04x.", mask );
    OSC_SHOUT;
    return -1;
  } 
  o->ddf.vermode &= ~(ddft)mask;
  o->ddf.vermode |= (ddft)mask & (ddft)mode; 
  sprintf( s, "VERMODE=%c", o->ddf.vermode );
  osc_write_strn( o, s, 9 ); on_error_return( -1 );
  osc_check_returncode( o, "osc_set_vermode: VERMODE" ); on_error_return( -1 );
  return 0;
}


/* --- trigger settings ------------------------------------------------- */

/**
 * Set Holdoff time
 *
 * o    	OSC struct
 * t		holdoff time, 0x00 is shortest, 0xff is longest
 * return: 0 = success; -1 = error
 */
int osc_set_holdoff( OSC * o, int t )
{
  char s[10];

  clear_error();
  if ( t < 0 || t > 0xff )
  {
    sprintf( o->error, "Holdoff must be in [0x00..0xff], not %#04x.", t );
    OSC_SHOUT;
    return -1;
  } 
  o->ddf.hold_off = t;
  sprintf( s, "HOLDOFF=%c", t );
  osc_write_strn( o, s, 9 ); on_error_return( -1 );
  osc_check_returncode( o, "osc_set_holdoff: HOLDOFF" ); on_error_return( -1 );
  return 0;
}


/**
 * Reset trigger.
 * Single mode storage: reset and wait for trigger.
 * Storage: Restart acquisition now.
 * Analog: start trigger now.
 *
 * o		OSC struct
 * return: 0 = success; -1 = error
 */
int osc_reset( OSC * o )
{
  osc_write_str( o, "RES" ); on_error_return( -1 );
  osc_check_returncode( o, "osc_reset: RES" ); on_error_return( -1 );
  return 0;
}


/**
 * set trigger mode.
 * mode comprises: OSC_TRI_...        NFL, PP, NORM
 * coupling:       OSC_TRI_COUPL_...  AC, DC, HF
 *
 * o		OSC struct
 * mode		desired mode, -1 = don't change
 * coupl	coupling, -1 = don't change
 * return: 0 = success; -1 = error
 */
int osc_set_trigger( OSC * o, int mode, int coupl )
{
  char s[7];

  clear_error();
  if ( mode & OSC_TRI_COUPL )
  {
    sprintf( o->error, "Illegal trigger mode %#04x.", mode );
    OSC_SHOUT;
    return -1;
  }
  if ( coupl & ~OSC_TRI_COUPL )
  {
    sprintf( o->error, "Illegal coupling mode %#04x.", mode );
    OSC_SHOUT;
    return -1;
  }
  if ( mode >= 0 )
    o->ddf.trig = ( o->ddf.trig & OSC_TRI_COUPL ) | (ddft)mode;
  if ( coupl >= 0 )
    o->ddf.trig = ( o->ddf.trig & ~OSC_TRI_COUPL ) | (ddft)coupl;
  if ( mode >= 0 || coupl >= 0 )
  {
    sprintf( s, "TRIG=%c", o->ddf.trig );
    osc_write_strn( o, s, 6 ); on_error_return( -1 );
    osc_check_returncode( o, "osc_set_trigger: TRIG" ); on_error_return( -1 );
  }
  return 0;
}


/**
 * Find out whether scope triggers.
 * States are:
 *   TR_YES  scope triggers
 *   TR_NO   socee triggers not
 *   TR_WAIT scope in ingle modus and waits or acquisition not complete
 *   -1	     error
 *
 * o		scope handle
 * fast		1 = without use of 200 ms monoflop
 * return: trigger status 
 */
int osc_get_trgsta( OSC * o, int fast )
{
  char b;

  clear_error();
  if ( fast )
  {
    osc_write_str( o, "TRGSTA" ); on_error_return(-1);
    osc_check_returncode( o, "osc_get_trgsta: TRGSTA" ); on_error_return(-1);
  }
  osc_write_str( o, "TRGSTA?" ); on_error_return(-1);
  osc_read_str( o, "TRGSTA:", &b, 1 ); on_error_return(-1);
  if ( b == 0 || b == '0' )
    return TR_NO;
  if ( b == 1 || b == '1' )
    return TR_YES;
  if ( b == 2 || b == '2' )
    return TR_WAIT;
  sprintf(  o->error, "Got illegal trigger state %#04x.", (int)b );
  OSC_SHOUT;
  return -1;
}


/**
 * Set pretrigger
 *
 * pt	one of OSC_STR_PRET_..  m75, m50, m25, 0, 25, 50, 75, 100
 * o		scope handle
 * return: 0 = success, -1 = error
 */
int osc_set_pretrig( OSC * o, int pt )
{
  char s[10];

  clear_error();
  if ( pt & ~OSC_STR_PRET )
  {
    sprintf( o->error, "Illegal pretrigger setting %#04x.", pt );
    OSC_SHOUT;
    return -1;
  } 
  o->ddf.strmode &= ~OSC_STR_PRET;
  o->ddf.strmode |= (ddft)pt;
  sprintf( s, "STRMODE=%c", o->ddf.strmode );
  osc_write_strn( o, s, 9 ); on_error_return( -1 );
  osc_check_returncode( o, "osc_set_pretrig: STRMODE" ); on_error_return( -1 );
  return 0;
}

/* --- reference wave --------------------------------------------------- */

/* --- Hold waveform ---------------------------------------------------- */

/**
 * Set hold waveform.
 *
 * o    	OSC struct
 * x		1 = hold, 0 = no hold
 * return: 0 = success, -1 = error
 */
int osc_set_hold( OSC * o, int x )
{
  char s[9];

  sprintf( s, "HLDWFM=%d", x?1:0 );
  osc_write_str( o, s ); on_error_return(-1);
  osc_check_returncode( o, "osc_set_hold: HLDWFM" ); on_error_return(-1);
  return 0;
}


/**
 * Get hold waveform.
 *
 * o    	OSC struct
 * return: 1 = hold, 0 = no hold, -1 = error
 */
int osc_get_hold( OSC * o )
/* not in ddf ??? */
{
  char c;

  osc_write_str( o, "HLDWFM?" ); on_error_return( -1 );
  osc_read_str( o, "HLDWFM:", &c, 1 ); on_error_return( -1 );
  switch ( c )
  {
    case '0': return 0;
    case '1': return 1;
    default:
      sprintf( o->error, "osc_get_hold got illegal answer %#04x.", 
               (unsigned int)c );
      OSC_SHOUT;
      return -1;
  }
}


