/**
 * HMsweep - software to measure frequnecy responses with the sound card and a
 * Hameg digital storage oscilloscope
 *         
 * Copyright (C) 2007 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * LSto@opppf.de
 * http://opppf.de/HMsweep/
 */
#include <math.h>
#include "sinegen.h"


#define SINEGEN_BUFLEN 44100

#if __BYTE_ORDER == __BIG_ENDIAN
# define LOCAL_PCM_FORMAT SND_PCM_FORMAT_S16_BE
#else
# define LOCAL_PCM_FORMAT SND_PCM_FORMAT_S16_LE
#endif


/* debugging */
#define debugc(x) fprintf( stderr, "%s=%c\n", #x, (x) )
#define debugi(x) fprintf( stderr, "%s=%d\n", #x, (x) )
#define debugd(x) fprintf( stderr, "%s=%f\n", #x, (x) )


/**
 * open the alsa device
 * 
 * alsa		alsa device handle
 * device	alsa device as string
 */
void sinegen_alsa_open( snd_pcm_t ** alsa, const char * device )
{
  int err;

  if ( ( err = snd_pcm_open( alsa,
                             device, 
                             SND_PCM_STREAM_PLAYBACK, 
	  		     0 ) ) < 0) 
  {
    fprintf( stderr, "Playback open error: %s\n", snd_strerror( err ) );
    exit( 1 );
  }
  if ( (err = snd_pcm_set_params( *alsa, 
                                  LOCAL_PCM_FORMAT,
                                  SND_PCM_ACCESS_RW_INTERLEAVED, 
                                  1, 
                                  44100, 
                                  1, 
                                  1000000 ) ) < 0 ) 
  {
    fprintf( stderr, "Playback open error: %s\n", snd_strerror( err ) );
    exit( 1 );
  }
}


/** 
 * close alsa device.
 * 
 * alsa		alsa device handle
 */
void sinegen_alsa_close( snd_pcm_t * alsa )
{
  snd_pcm_close( alsa );
}


/**
 * Fill buffer with sine wave.
 *
 * f	desired frequency
 * p	phase of oscillator
 * b	buffer
 * return: new phase of oscillator
 */
double sinegen_fill_buffer( double f, double p, signed short int * b )
{
  int i;

  for ( i = 0; i < SINEGEN_BUFLEN; i++ )
  {
    p += f / 44100.0;
    b[i] = (signed short int)floor( 0.5 + 32000 * cos( 2 * M_PI * p ) );
  }
  return p - floor( p );
}


/**
 * thread functioin playing sound.
 */
void * sinegen_thread( void * arg )
{
  sinegen_alsa_t * sg = (sinegen_alsa_t *)arg;
  snd_pcm_sframes_t frames;
  signed short int buf[SINEGEN_BUFLEN];
  double phase = 0;

  pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, NULL );
  pthread_setcanceltype( PTHREAD_CANCEL_DEFERRED, NULL );

  while ( 1 )
  {
    phase = sinegen_fill_buffer( sg->f, phase, buf );
    frames = snd_pcm_writei( sg->alsa, buf, SINEGEN_BUFLEN );
    if ( frames < 0 )
      frames = snd_pcm_recover( sg->alsa, frames, 0 );
    if ( frames < 0 ) 
      fprintf( stderr, "snd_pcm_writei failed: %s\n", snd_strerror( frames ) );
    if ( frames > 0 && frames < SINEGEN_BUFLEN )
      fprintf( stderr, "Short write (expected %li, wrote %li)\n", 
               (long)SINEGEN_BUFLEN, frames );
    pthread_testcancel();
  }
  return NULL; /* never reached */
}



/* **********************************************************************
 * api functions
 * ********************************************************************** */


/**
 * Start sine generator with frequency f.
 *
 * alsa		alsa deviece for output
 * f		frequency in Hz
 * return: sinegen handle, NULL on error
 */
sinegen_alsa_t * sinegen_alsa_start( const char * device, double f )
{
  sinegen_alsa_t * sg = malloc( sizeof(sinegen_alsa_t) );

  if ( f > 22050.0 )
  {
    if ( f - 22050.0 < 0.1 ) /* it might be due to rounding */
      f = 22050.0;
    else
    {
      fprintf( stderr, "%f is above Nyquist-frequency.\n", f );
      return NULL;
    } 
  }
  sg->f = f;
  sinegen_alsa_open( &(sg->alsa), device );
  pthread_create( &(sg->thread), NULL, &sinegen_thread, sg );
  return sg;
}


/**
 * Stop sine generator.
 *
 * sg	singe generator
 */
void sinegen_alsa_stop( sinegen_alsa_t * sg )
{
  pthread_cancel( sg->thread );
  pthread_join( sg->thread, NULL );
  sinegen_alsa_close( sg->alsa );
  free( sg );
}



